/* 
 * xnetload.c: multiple xload, using rwho statistics
 * $Date: 1995/02/14 03:28:02 $ $Revision: 1.16 $
 *
 * Copyright 1989 University of Waterloo
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of UW not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  UW makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * UW DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL UW
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Mike Berkley, jmberkley@watnext.waterloo.edu, 1989
 *
 */

/*
 * xnetload uses the Load widget load proc to get the local
 * machine load.  This means that xnetload needs to be setgid
 * kmem.  If you do not want this, then define NOLOCAL when compiling.
 */

#include <stdio.h>
#ifndef __STRICT_BSD__
#include <stdlib.h>
#endif
#include <netdb.h>
#include <ctype.h>
#include <sys/param.h>
#include <X11/Xos.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/StripChart.h>
#include <X11/Xmu/SysUtil.h>
#include "patchlevel.h"

void GetStatus();
extern int getload();         /* gets rwho load */
#if defined(sun) || defined(__STRICT_BSD__)
extern char *malloc();
#endif
extern char *strtok();
extern int gethostname();
static Widget makestatuswidget(), makeremotestatus(), makesubform();
static void modifyloadcall();
static int  matchstr();
static void updatelabel();
static void usage();

#define DOWNMSG      " DOWN"
#define UPMSG        "     "	/* has to have same length as DOWNMSG */
#define BIGNUM (99999)		/* ~infinite number of load widgets */

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif

/* arguments to Form widget */
static Arg Formargs[] = {
    { XtNfromVert, (XtArgVal) NULL },       /* Chain from .value wid */
    { XtNdefaultDistance, (XtArgVal) 1 },   /* default spacing */
    { XtNresizable, (XtArgVal) TRUE },      /* allow resize */
};

/* Arguments to load windows */
static Arg Statusargs[] = {
    { XtNfromHoriz, (XtArgVal) NULL },      /* Chain from .value wid */
    { XtNresizable, (XtArgVal) TRUE },      /* allow resize */
    { XtNborderWidth, (XtArgVal) 0 },       /* 0 border for load wid  */
    { XtNlabel, (XtArgVal) NULL },          /* Load title */
};

/* Application specific resources - turn off local load display */
static struct RsrcData {
  String Configuration;
  Boolean Localload;
} rsrcData;
static XtResource app_resources[] = {
    { "localload", "LocalLoad", XtRBoolean, sizeof(Boolean),
	XtOffsetOf(struct RsrcData, Localload), XtRString,
/* if NOLOCAL is defined, then no local is displayed, by default */
#ifdef NOLOCAL
"off"
#else
"on"
#endif
},
  /* configuration of loads - horizontal, vertical or mXn/mxn */
  { "configuration", "Configuration", XtRString, sizeof(String),
      XtOffsetOf(struct RsrcData, Configuration), XtRImmediate,
      "horizontal" },
};

/* command line options for local load display and configuration */
/* If compiled NOLOCAL, then off is default, but on still allowed */
static XrmOptionDescRec app_options[] = {
    { "-local", "localload", XrmoptionNoArg, "on" },
    { "-nolocal", "localload", XrmoptionNoArg, "off" },
    { "-configuration", "configuration", XrmoptionSepArg, "vertical"},
};

int
main(argc,argv)
int argc;
char *argv[];
{
    int currarg;             /* arg counter to walk through hosts */
    int maincount,subcount,count;
    String maindirection,subdirection;
    static Widget toplevel, mainform, subform, statuswin;

    /* The toolkit tries to open your display if you use -help */
    /* so let's do it the other way */
    if((argc > 1) && !strncmp(argv[1],"-help",5)) {
        usage();
        exit(0);
    }

    /* Initialize the top level */
    toplevel = XtInitialize(argv[0], "XNetload",
                            app_options, XtNumber(app_options), &argc, argv);

    /* Get local vs nolocal, configuration */
    XtGetApplicationResources( toplevel, (XtPointer)&rsrcData,
			      app_resources, XtNumber(app_resources),
			      NULL, ZERO);
      
    /* Parse configuration */
    if(strncmp(rsrcData.Configuration,"horiz",5) == 0) {
	subcount = BIGNUM;	/* ~ inf num of widgets horizontally */
	maincount = 1;
    }
    else if(strncmp(rsrcData.Configuration,"vert",4) == 0) {
	maincount = BIGNUM;	/* ~ inf num of widgets horizontally */
	subcount = 1;
    }
    else {
	/* format is 4x5 for 4 columns, 5 rows */
	strtok(rsrcData.Configuration,"xX");
	if(rsrcData.Configuration[0] == '*')
	  subcount = BIGNUM;
	else if(! isdigit(rsrcData.Configuration[0])) {
	    usage();
	    exit(-1);
	}
	else if((subcount = atoi(rsrcData.Configuration)) == 0)
	  subcount = 1;
	
	rsrcData.Configuration = strtok((char *)NULL,"xX");
	if(rsrcData.Configuration[0] == '*')
	  maincount = BIGNUM;
	else if(! isdigit(rsrcData.Configuration[0])) {
	    usage();
	    exit(-1);
	}
	else if((maincount = atoi(rsrcData.Configuration)) == 0)
	  maincount = 1;
    }

    if(subcount >= maincount) {
	maindirection = XtNfromVert;
	subdirection = XtNfromHoriz;
    }
    /* need to swap so that bigger count is the sub count */
    /* This lowers the number of form widgets */
    else {
	maindirection = XtNfromHoriz;
	subdirection = XtNfromVert;
	/* swap counts */
	count = subcount;subcount = maincount;maincount = count;
    }
    /* adjust subcount so that we don't always get one row if we use Nx* */
    if(subcount == BIGNUM) {
	subcount = argc/maincount;
	if(maincount * subcount != argc)
	  subcount++;
    }

    /* Create a Form widget to contain all of the Loads */
    mainform =
      XtCreateManagedWidget("outerform",
                            formWidgetClass,
                            toplevel,
                            Formargs,
                            XtNumber(Formargs));

    /* Make the first sub-form widget, to contain one row/col */
    subform = makesubform(mainform,NULL,maindirection);

    /*
     * load widget reads kmem for local load.  Define NOLOCAL if you
     * don't want this.
     */
    count = 0;
    if(rsrcData.Localload) {
        /* Get and set label name */
        String labelname = malloc(BUFSIZ*sizeof(char));
        gethostname(labelname,BUFSIZ);
	strtok(labelname,".");

        /* Create the local load widget */
        statuswin =
	  XtParent(makestatuswidget(labelname,subform,(Widget)NULL,
				    "local",subdirection));
	count++;		/* So local added into subform count */
    }

    /* Remote Loads */
    /* currarg initialized above */
    for(currarg = 1;(currarg < argc)&&(maincount > 0);maincount--) {
	for(;(currarg < argc)&&(count < subcount);count++,currarg++) {
	    if(count > 0) {
		/* Chain to first load */
		statuswin = makeremotestatus(argv[currarg],
					     subform,
					     statuswin,
					     subdirection);
	    }
	    else {
		/* chain to left/top edge of subform */
		statuswin = makeremotestatus(argv[currarg],
					     subform,
					     NULL,
					     subdirection);
	    }
	}
	count = 0;
	/* make another subform if there are more machines */
	if((currarg < argc) && (maincount > 1))
	  subform = makesubform(mainform,subform,maindirection);
    }

    /* if there are no loads to display, then should stop   */
    /* Note, statuswin is static, guaranteed to start at 0, */
    /* so it's non-zero only if a load widget was created   */
    if(!statuswin) {
        fprintf(stderr,"xnetload: No machines to display\n");
        usage();
        exit(0);
    }

    /* draw the widgets, and look for new events */
    XtRealizeWidget(toplevel);
    XtMainLoop();
}

/*
 * Make a sub form widget
 */
static Widget
makesubform(formwin,lastwin,direction)
Widget formwin,lastwin;
String direction;
{
    /* set up chain to previous widget */
    XtSetArg(Formargs[0],direction, (XtArgVal)lastwin);

    /* Create the local form */
    return(XtCreateManagedWidget("subform",
				 formWidgetClass,
				 formwin,
				 Formargs,
				 XtNumber(Formargs)));
}

/*
 * Make a status widget
 */
static Widget
makestatuswidget(label,formwin,lastwin,name,direction)
char *label;
Widget formwin,lastwin;
String name,direction;
{
    Widget load, pane, label_wid;
    Arg	   args[1];
    void   GetLoadPoint();	/* from the xload program */

    XtSetArg(Statusargs[XtNumber(Statusargs)-1],
	     XtNlabel,(XtArgVal)label);
    /* set up chain to previous Load widget */
    XtSetArg(Statusargs[0],direction,(XtArgVal)lastwin);

    pane = XtCreateManagedWidget( "paned", panedWidgetClass,
				  formwin,
				  Statusargs,
				  XtNumber(Statusargs) );

    XtSetArg( args[0], XtNlabel, label );
    label_wid = XtCreateManagedWidget( "label", labelWidgetClass,
				       pane, args, ONE );

    /* Create the load widget */
    load = XtCreateManagedWidget( name,
				  stripChartWidgetClass,
				  pane,
				  NULL, ZERO );
    XtAddCallback(load, XtNgetValue, GetLoadPoint, NULL);
    return load;
}

/*
 * Make a remote status widget
 */
static Widget
makeremotestatus(hostname,formwin,lastwin,direction)
String hostname;
Widget formwin,lastwin;
String direction;
{
    String labelname;		/* label, could be "machine DOWN" */
    struct hostent *host;	/* for gethostbyname() */
    Widget remotestatus;

    /* translate command arg to a hostname */
    host = gethostbyname(hostname);
    /* if host == NULL, then hostname is junk for some reason */
    if(!host) 
      return(NULL);

    /* get rid of domain's in fully specified host names */
    strtok(host->h_name,".");
    /* allocate memory for name and downmsg for label */
    if(!(labelname=malloc((strlen(host->h_name)
			   +strlen(DOWNMSG)+1)*sizeof(char))
	 )) {
	perror("xnetload: Ran out of memory for labelname");
	exit(1);
    }
    /* set label name */
    strcpy(labelname,host->h_name);
    strcat(labelname,UPMSG);
    remotestatus = makestatuswidget(labelname,formwin,lastwin,
				    "remote",direction);

    free(labelname);
    modifyloadcall(remotestatus, host->h_name);
    return(XtParent(remotestatus));
}
    

/*
 * replace standard load widget status call with one that looks at rwho
 */
static void
modifyloadcall(statuswin,name)
Widget statuswin;
String name;
{
    String hostname;

    /* allocate memory for hostname */
    if(!(hostname=
	 malloc((strlen(name)+1)*sizeof(char)))) {
	perror("xnetload: Ran out of memory for name");
	exit(1);
    }
    /* Set up hostname, and load callback (GetStatus()) */
    strcpy(hostname,name);
    XtRemoveAllCallbacks(statuswin, XtNgetValue);
    XtAddCallback(statuswin,XtNgetValue,
		  GetStatus,(caddr_t)hostname);
}

/* GetStatus() global vars */
static char Label[MAXHOSTNAMELEN + sizeof(DOWNMSG)];
static Arg Labelarg[1];

/*
 * Load callback function
 *    - calls getload to determine current load, then updates labels
 */
void
GetStatus(w,closure,call_data)
Widget w;                        /* load widget */
caddr_t closure;                 /* name of host, char*  */
caddr_t call_data;               /* load, float*  */
{
    int load;			/* current load for this machine */
    String ltext;
    Widget load_parent, label_wid;

    load_parent = XtParent( w );
    label_wid   = XtNameToWidget( load_parent, "label" );

    /* get current label for widget */
    XtSetArg( Labelarg[0], XtNlabel,( XtArgVal)&ltext);
    XtGetValues( label_wid, Labelarg, ONE);
    strcpy(Label, ltext);

    /* if load is less than zero, then machine is down or something */
    load = getload(closure);
    if( load < 0 )
    {
        *(double *)call_data = 0.0;
        /* update label with DOWN message */
        if( !matchstr(Label,DOWNMSG) )
	{
	    updatelabel( label_wid, (char *)closure, DOWNMSG );
	}
    }
    else			/* otherwise we found a good load, so use it */
    {
        *(double *)call_data = (double)load/100.0;
        /* hate to do this everytime, but we have to check */
        if( matchstr(Label,DOWNMSG) )
	{
	    updatelabel( label_wid, (char *)closure, UPMSG);
	}
        return;
    }
}

/*
 * look for pat in str - exact match, no wild cards here
 */
static int
matchstr(str,pat)
register char *str, *pat;
{
    while(str = index(str,*pat))
      if(! strcmp(str++,pat))
        return(TRUE);
    return(FALSE);
}

/*
 * replace label with str and msg
 */
static void
updatelabel(w,str,msg)
Widget w;
register char *str, *msg;
{
    sprintf(Label,"%.*s%.*s", MAXHOSTNAMELEN, str, sizeof(DOWNMSG), msg);
    XtSetArg(Labelarg[0],XtNlabel, (XtArgVal)Label);
    XtSetValues(w,Labelarg, ONE);
}

static void
usage()
{
#ifdef NOLOCAL
    fprintf(stderr,"usage:\n xnetload [-configuration mXn] [-nolocal] [machine] ...\n");
#else
    fprintf(stderr,"usage:\n xnetload [-configuration mXn] [-local] [machine] ...\n");
#endif
}
